/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// draw.c -- this is the only file outside the refresh that touches the
// vid buffer

#include "quakedef.h"

#include "zlib/png.h" //rww

extern qboolean g_canUseTexComp; //rww
extern unsigned char d_15to8table[65536];
extern cvar_t crosshair, cl_crossx, cl_crossy, crosshaircolor;

qboolean GL_UploadPNG(char *filename, int tex, qboolean mips);
int hr_plaque = 0; //rww
int hr_conback = 0; //rww
int hr_sun = 0; //rww
int hr_sunring = 0; //rww
int hr_blood1 = 0; //rww
int hr_blood2 = 0; //rww
int hr_blood3 = 0; //rww
int hr_blood4 = 0; //rww
int hr_water = 0; //rww
int hr_water_n = 0; //rww
int hr_water2 = 0; //rww
int hr_water2_n = 0; //rww
int hr_leaf = 0; //rww

cvar_t		gl_nobind = {"gl_nobind", "0"};
cvar_t		gl_max_size = {"gl_max_size", "1024"};
cvar_t		gl_picmip = {"gl_picmip", "0"};
cvar_t		gl_hq4x = {"gl_hq4x", "1"}; //rww
cvar_t		gl_normalMaps = {"gl_normalMaps", "1", true}; //rww
cvar_t		gl_texCompression = {"gl_texCompression", "0", true}; //rww
cvar_t		gl_bloom = {"gl_bloom", "1", true}; //rww
cvar_t		gl_dynamicMapLights = {"gl_dynamicMapLights", "1", true}; //rww
cvar_t		gl_ambientShadows = {"gl_ambientShadows", "1", true}; //rww
cvar_t		gl_qualityFoliage = {"gl_qualityFoliage", "1", true}; //rww
cvar_t		gl_lowQualityBloom = {"gl_lowQualityBloom", "0", true}; //rww
cvar_t		gl_lightCounts = {"gl_lightCounts", "0"}; //rww
cvar_t		gl_twoSidedStencil = {"gl_twoSidedStencil", "1"}; //rww
cvar_t		gl_forceWideAspect = {"gl_forceWideAspect", "0"}; //rww
cvar_t		gl_surfSpriteDensity = {"gl_surfSpriteDensity", "0.005"}; //rww
cvar_t		gl_specExp = {"gl_specExp", "12.0"}; //rww
cvar_t		gl_showTris = {"gl_showTris", "0"}; //rww
cvar_t		gl_lightScissor = {"gl_lightScissor", "1"}; //rww

byte		*draw_chars;				// 8*8 graphic characters
qpic_t		*draw_disc;
qpic_t		*draw_backtile;

int			translate_texture;
int			char_texture;
int			cs_texture; // crosshair texture

static byte cs_data[64] = {
	0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xff,
	0xfe, 0xff, 0xfe, 0xff, 0xfe, 0xff, 0xfe, 0xff,
	0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xfe, 0xff, 0xff, 0xff, 0xff,
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};


typedef struct
{
	int		texnum;
	float	sl, tl, sh, th;
} glpic_t;

byte		conback_buffer[sizeof(qpic_t) + sizeof(glpic_t)];
qpic_t		*conback = (qpic_t *)&conback_buffer;

int		gl_lightmap_format = 4;
int		gl_solid_format = 3;
int		gl_alpha_format = 4;

int		gl_filter_min = GL_LINEAR_MIPMAP_LINEAR; //rww GL_LINEAR_MIPMAP_NEAREST;
int		gl_filter_max = GL_LINEAR;


int		texels;

#define	MAX_GLTEXTURES	1024
gltexture_t	gltextures[MAX_GLTEXTURES];
int			numgltextures;

void GL_Bind (int texnum)
{
	if (gl_nobind.value)
		texnum = char_texture;
	if (currenttexture == texnum)
		return;
	currenttexture = texnum;
#ifdef _WIN32
	bindTexFunc (GL_TEXTURE_2D, texnum);
#else
	glBindTexture (GL_TEXTURE_2D, texnum);
#endif
}


/*
=============================================================================

  scrap allocation

  Allocate all the little status bar obejcts into a single texture
  to crutch up stupid hardware / drivers

=============================================================================
*/

#define	MAX_SCRAPS		1
#define	BLOCK_WIDTH		256
#define	BLOCK_HEIGHT	256

int			scrap_allocated[MAX_SCRAPS][BLOCK_WIDTH];
byte		scrap_texels[MAX_SCRAPS][BLOCK_WIDTH*BLOCK_HEIGHT*4];
qboolean	scrap_dirty;
int			scrap_texnum;

// returns a texture number and the position inside it
int Scrap_AllocBlock (int w, int h, int *x, int *y)
{
	int		i, j;
	int		best, best2;
	int		texnum;

	for (texnum=0 ; texnum<MAX_SCRAPS ; texnum++)
	{
		best = BLOCK_HEIGHT;

		for (i=0 ; i<BLOCK_WIDTH-w ; i++)
		{
			best2 = 0;

			for (j=0 ; j<w ; j++)
			{
				if (scrap_allocated[texnum][i+j] >= best)
					break;
				if (scrap_allocated[texnum][i+j] > best2)
					best2 = scrap_allocated[texnum][i+j];
			}
			if (j == w)
			{	// this is a valid spot
				*x = i;
				*y = best = best2;
			}
		}

		if (best + h > BLOCK_HEIGHT)
			continue;

		for (i=0 ; i<w ; i++)
			scrap_allocated[texnum][*x + i] = best + h;

		return texnum;
	}

	Sys_Error ("Scrap_AllocBlock: full");
	return 0;
}

int	scrap_uploads;

void Scrap_Upload (void)
{
	scrap_uploads++;
	GL_Bind(scrap_texnum);
	GL_Upload8 (NULL, scrap_texels[0], BLOCK_WIDTH, BLOCK_HEIGHT, false, true);
	scrap_dirty = false;
}

//=============================================================================
/* Support Routines */

typedef struct cachepic_s
{
	char		name[MAX_QPATH];
	qpic_t		pic;
	byte		padding[32];	// for appended glpic
} cachepic_t;

#define	MAX_CACHED_PICS		128
cachepic_t	menu_cachepics[MAX_CACHED_PICS];
int			menu_numcachepics;

byte		menuplyr_pixels[4096];

int		pic_texels;
int		pic_count;

qpic_t *Draw_PicFromWad (char *name)
{
	qpic_t	*p;
	glpic_t	*gl;

	p = W_GetLumpName (name);
	gl = (glpic_t *)p->data;

	// load little ones into the scrap
	if (p->width < 64 && p->height < 64)
	{
		int		x, y;
		int		i, j, k;
		int		texnum;

		texnum = Scrap_AllocBlock (p->width, p->height, &x, &y);
		scrap_dirty = true;
		k = 0;
		for (i=0 ; i<p->height ; i++)
			for (j=0 ; j<p->width ; j++, k++)
				scrap_texels[texnum][(y+i)*BLOCK_WIDTH + x + j] = p->data[k];
		texnum += scrap_texnum;
		gl->texnum = texnum;
		gl->sl = (x+0.01)/(float)BLOCK_WIDTH;
		gl->sh = (x+p->width-0.01)/(float)BLOCK_WIDTH;
		gl->tl = (y+0.01)/(float)BLOCK_WIDTH;
		gl->th = (y+p->height-0.01)/(float)BLOCK_WIDTH;

		pic_count++;
		pic_texels += p->width*p->height;
	}
	else
	{
		gl->texnum = GL_LoadPicTexture (p);
		gl->sl = 0;
		gl->sh = 1;
		gl->tl = 0;
		gl->th = 1;
	}
	return p;
}


/*
================
Draw_CachePic
================
*/
qboolean g_loadingPicTex = false; //rww
qpic_t	*Draw_CachePic (char *path)
{
	cachepic_t	*pic;
	int			i;
	qpic_t		*dat;
	glpic_t		*gl;

	for (pic=menu_cachepics, i=0 ; i<menu_numcachepics ; pic++, i++)
		if (!strcmp (path, pic->name))
			return &pic->pic;

	if (menu_numcachepics == MAX_CACHED_PICS)
		Sys_Error ("menu_numcachepics == MAX_CACHED_PICS");
	menu_numcachepics++;
	strcpy (pic->name, path);

//
// load the pic from disk
//
	dat = (qpic_t *)COM_LoadTempFile (path);	
	if (!dat)
		Sys_Error ("Draw_CachePic: failed to load %s", path);
	SwapPic (dat);

	// HACK HACK HACK --- we need to keep the bytes for
	// the translatable player picture just for the menu
	// configuration dialog
	if (!strcmp (path, "gfx/menuplyr.lmp"))
		memcpy (menuplyr_pixels, dat->data, dat->width*dat->height);

	pic->pic.width = dat->width;
	pic->pic.height = dat->height;

//	g_loadingPicTex = true; //rww
	gl = (glpic_t *)pic->pic.data;
	gl->texnum = GL_LoadPicTexture (dat);
//	g_loadingPicTex = false; //rww
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;

	return &pic->pic;
}


void Draw_CharToConback (int num, byte *dest)
{
	int		row, col;
	byte	*source;
	int		drawline;
	int		x;

	row = num>>4;
	col = num&15;
	source = draw_chars + (row<<10) + (col<<3);

	drawline = 8;

	while (drawline--)
	{
		for (x=0 ; x<8 ; x++)
			if (source[x] != 255)
				dest[x] = /*0x60 +*/ source[x];
		source += 128;
		dest += 320;
	}

}

typedef struct
{
	char *name;
	int	minimize, maximize;
} glmode_t;

glmode_t modes[] = {
	{"GL_NEAREST", GL_NEAREST, GL_NEAREST},
	{"GL_LINEAR", GL_LINEAR, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_NEAREST", GL_NEAREST_MIPMAP_NEAREST, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_NEAREST", GL_LINEAR_MIPMAP_NEAREST, GL_LINEAR},
	{"GL_NEAREST_MIPMAP_LINEAR", GL_NEAREST_MIPMAP_LINEAR, GL_NEAREST},
	{"GL_LINEAR_MIPMAP_LINEAR", GL_LINEAR_MIPMAP_LINEAR, GL_LINEAR}
};

/*
===============
Draw_TextureMode_f
===============
*/
void Draw_TextureMode_f (void)
{
	int		i;
	gltexture_t	*glt;

	if (Cmd_Argc() == 1)
	{
		for (i=0 ; i< 6 ; i++)
			if (gl_filter_min == modes[i].minimize)
			{
				Con_Printf ("%s\n", modes[i].name);
				return;
			}
		Con_Printf ("current filter is unknown???\n");
		return;
	}

	for (i=0 ; i< 6 ; i++)
	{
		if (!Q_strcasecmp (modes[i].name, Cmd_Argv(1) ) )
			break;
	}
	if (i == 6)
	{
		Con_Printf ("bad filter name\n");
		return;
	}

	gl_filter_min = modes[i].minimize;
	gl_filter_max = modes[i].maximize;

	// change all the existing mipmap texture objects
	for (i=0, glt=gltextures ; i<numgltextures ; i++, glt++)
	{
		if (glt->mipmap)
		{
			GL_Bind (glt->texnum);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
			glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
		}
	}
}

//rww
void Draw_LoadHiRes(void)
{
	hr_plaque = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/plaque.png", hr_plaque, false))
	{
		hr_plaque = 0;
		texture_extension_number--;
	}

	hr_conback = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/conback.png", hr_conback, false))
	{
		hr_conback = 0;
		texture_extension_number--;
	}

	hr_sun = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/sun.png", hr_sun, false))
	{
		hr_sun = 0;
		texture_extension_number--;
	}
	hr_sunring = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/sunring.png", hr_sunring, false))
	{
		hr_sunring = 0;
		texture_extension_number--;
	}

	hr_blood1 = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/blood1.png", hr_blood1, true))
	{
		hr_blood1 = 0;
		texture_extension_number--;
	}
	else
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}
	hr_blood2 = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/blood2.png", hr_blood2, true))
	{
		hr_blood2 = 0;
		texture_extension_number--;
	}
	else
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}
	hr_blood3 = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/blood3.png", hr_blood3, true))
	{
		hr_blood3 = 0;
		texture_extension_number--;
	}
	else
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}

	hr_blood4 = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/blood4.png", hr_blood4, true))
	{
		hr_blood4 = 0;
		texture_extension_number--;
	}

	/*
	hr_water = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/water.png", hr_water, true))
	{
		hr_water = 0;
		texture_extension_number--;
	}
	*/
	hr_water_n = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/water_n.png", hr_water_n, true))
	{
		hr_water_n = 0;
		texture_extension_number--;
	}

	hr_water2 = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/water2.png", hr_water2, true))
	{
		hr_water2 = 0;
		texture_extension_number--;
	}
	hr_water = hr_water2;
	hr_water2_n = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/other/water2_n.png", hr_water2_n, true))
	{
		hr_water2_n = 0;
		texture_extension_number--;
	}

	hr_leaf = texture_extension_number;
	texture_extension_number++;
	if (!GL_UploadPNG("qdix/assets/textures/leaf01.png", hr_leaf, true))
	{
		hr_leaf = 0;
		texture_extension_number--;
	}

	GL_Bind(0);
}

/*
===============
Draw_Init
===============
*/
void Draw_Init (void)
{
	int		i;
	qpic_t	*cb;
	byte	*dest;
	int		x;
	char	ver[40];
	glpic_t	*gl;
	int start;
	byte    *ncdata;

	Cvar_RegisterVariable (&gl_nobind);
	Cvar_RegisterVariable (&gl_max_size);
	Cvar_RegisterVariable (&gl_picmip);
	Cvar_RegisterVariable (&gl_hq4x); //rww
	Cvar_RegisterVariable (&gl_normalMaps); //rww
	Cvar_RegisterVariable (&gl_texCompression); //rww
	Cvar_RegisterVariable (&gl_bloom); //rww
	Cvar_RegisterVariable (&gl_dynamicMapLights); //rww
	Cvar_RegisterVariable (&gl_ambientShadows); //rww
	Cvar_RegisterVariable (&gl_qualityFoliage); //rww
	Cvar_RegisterVariable (&gl_lowQualityBloom); //rww
	Cvar_RegisterVariable (&gl_lightCounts); //rww
	Cvar_RegisterVariable (&gl_twoSidedStencil); //rww
	Cvar_RegisterVariable (&gl_forceWideAspect); //rww
	Cvar_RegisterVariable (&gl_surfSpriteDensity); //rww
	Cvar_RegisterVariable (&gl_specExp); //rww
	Cvar_RegisterVariable (&gl_showTris); //rww
	Cvar_RegisterVariable (&gl_lightScissor); //rww

	// 3dfx can only handle 256 wide textures
	if (!Q_strncasecmp ((char *)gl_renderer, "3dfx",4) ||
		!Q_strncasecmp ((char *)gl_renderer, "Mesa",4))
		Cvar_Set ("gl_max_size", "256");

	Cmd_AddCommand ("gl_texturemode", &Draw_TextureMode_f);

	// load the console background and the charset
	// by hand, because we need to write the version
	// string into the background before turning
	// it into a texture
	draw_chars = W_GetLumpName ("conchars");
	for (i=0 ; i<256*64 ; i++)
	{
		if (draw_chars[i] == 0)
		{
			draw_chars[i] = 255;	// proper transparent color
		}
	}

	// now turn them into textures
	char_texture = GL_LoadTexture ("charset", 128, 128, draw_chars, false, true, NULL);
//	Draw_CrosshairAdjust();
	cs_texture = GL_LoadTexture ("crosshair", 8, 8, cs_data, false, true, NULL);

	Draw_LoadHiRes();

	start = Hunk_LowMark ();

	cb = (qpic_t *)COM_LoadHunkFile ("gfx/conback.lmp");	
	if (!cb)
		Sys_Error ("Couldn't load gfx/conback.lmp");
	SwapPic (cb);

	sprintf (ver, "%4.2f", VERSION);
	dest = cb->data + 320 + 320*186 - 11 - 8*strlen(ver);
	for (x=0 ; x<strlen(ver) ; x++)
		Draw_CharToConback (ver[x], dest+(x<<3));

#if 0
	conback->width = vid.conwidth;
	conback->height = vid.conheight;

	// scale console to vid size
	dest = ncdata = Hunk_AllocName(vid.conwidth * vid.conheight, "conback");

	for (y=0 ; y<vid.conheight ; y++, dest += vid.conwidth)
	{
		src = cb->data + cb->width * (y*cb->height/vid.conheight);
		if (vid.conwidth == cb->width)
			memcpy (dest, src, vid.conwidth);
		else
		{
			f = 0;
			fstep = cb->width*0x10000/vid.conwidth;
			for (x=0 ; x<vid.conwidth ; x+=4)
			{
				dest[x] = src[f>>16];
				f += fstep;
				dest[x+1] = src[f>>16];
				f += fstep;
				dest[x+2] = src[f>>16];
				f += fstep;
				dest[x+3] = src[f>>16];
				f += fstep;
			}
		}
	}
#else
	conback->width = cb->width;
	conback->height = cb->height;
	ncdata = cb->data;
#endif
	
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	gl = (glpic_t *)conback->data;
	gl->texnum = GL_LoadTexture ("conback", conback->width, conback->height, ncdata, false, false, NULL);
	gl->sl = 0;
	gl->sh = 1;
	gl->tl = 0;
	gl->th = 1;
	conback->width = vid.conwidth;
	conback->height = vid.conheight;

	// free loaded console
	Hunk_FreeToLowMark (start);

	// save a texture slot for translated picture
	translate_texture = texture_extension_number++;

	// save slots for scraps
	scrap_texnum = texture_extension_number;
	texture_extension_number += MAX_SCRAPS;

	//
	// get the other pics we need
	//
	draw_disc = Draw_PicFromWad ("disc");
	draw_backtile = Draw_PicFromWad ("backtile");
}



/*
================
Draw_Character

Draws one 8*8 graphics character with 0 being transparent.
It can be clipped to the top of the screen to allow the console to be
smoothly scrolled off.
================
*/
void Draw_Character (int x, int y, int num)
{
	int				row, col;
	float			frow, fcol, size;

	if (num == 32)
		return;		// space

	num &= 255;
	
	if (y <= -8)
		return;			// totally off screen

	row = num>>4;
	col = num&15;

	if (row == 8)
	{ //rww - hack to eliminate pixel bleeding on certain character's top edges
		frow = row*0.0628;
	}
	else
	{
		frow = row*0.0625;
	}
	fcol = col*0.0625;
	size = 0.0625;

	GL_Bind (char_texture);

	/*
	//rww begin
	{
		const float backW = 0.6f;
		const float backH = 0.6f;

		glColor4f(0.0f, 0.0f, 0.0f, 1.0f);
		glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
		glBegin (GL_QUADS);
		glTexCoord2f (fcol, frow);
		glVertex2f (x-backW, y-backH);
		glTexCoord2f (fcol + size, frow);
		glVertex2f (x+8+backW, y-backH);
		glTexCoord2f (fcol + size, frow + size);
		glVertex2f (x+8+backW, y+8+backH);
		glTexCoord2f (fcol, frow + size);
		glVertex2f (x-backW, y+8+backH);
		glEnd ();
		glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	}
	//rww end
	*/

	glBegin (GL_QUADS);
	glTexCoord2f (fcol, frow);
	glVertex2f (x, y);
	glTexCoord2f (fcol + size, frow);
	glVertex2f (x+8, y);
	glTexCoord2f (fcol + size, frow + size);
	glVertex2f (x+8, y+8);
	glTexCoord2f (fcol, frow + size);
	glVertex2f (x, y+8);
	glEnd ();
}

extern int window_width, window_height;

//rww
float Draw_ScaleBasedOnResW(float f)
{
	return f*((float)vid.width/640.0f);
}

//rww
float Draw_ScaleBasedOnResH(float f)
{
	return f*((float)vid.height/480.0f);
}

//rww
void Draw_CharacterScaled (int x, int y, int num)
{
	int				row, col;
	float			frow, fcol, size;
	float			fX, fY, fW, fH;

	if (num == 32)
		return;		// space

	num &= 255;
	
	if (y <= -8)
		return;			// totally off screen

	row = num>>4;
	col = num&15;

	if (row == 8)
	{ //rww - hack to eliminate pixel bleeding on certain character's top edges
		frow = row*0.0628;
	}
	else
	{
		frow = row*0.0625;
	}
	fcol = col*0.0625;
	size = 0.0625;

	fX = Draw_ScaleBasedOnResW((float)x);
	fY = Draw_ScaleBasedOnResH((float)y);
	fW = Draw_ScaleBasedOnResW(8.0f);
	fH = Draw_ScaleBasedOnResH(8.0f);

	GL_Bind (char_texture);

	glBegin (GL_QUADS);
	glTexCoord2f (fcol, frow);
	glVertex2f (fX, fY);
	glTexCoord2f (fcol + size, frow);
	glVertex2f (fX+fW, fY);
	glTexCoord2f (fcol + size, frow + size);
	glVertex2f (fX+fW, fY+fH);
	glTexCoord2f (fcol, frow + size);
	glVertex2f (fX, fY+fH);
	glEnd ();
}

/*
================
Draw_String
================
*/
void Draw_String (int x, int y, char *str)
{
	while (*str)
	{
		Draw_Character (x, y, *str);
		str++;
		x += 8;
	}
}

/*
================
Draw_Alt_String
================
*/
void Draw_Alt_String (int x, int y, char *str)
{
	while (*str)
	{
		Draw_Character (x, y, (*str) | 0x80);
		str++;
		x += 8;
	}
}

void Draw_Crosshair(void)
{
	int x, y;
	extern vrect_t		scr_vrect;
	unsigned char *pColor;

	if (crosshair.value == 2) {
		x = scr_vrect.x + scr_vrect.width/2 - 3 + cl_crossx.value; 
		y = scr_vrect.y + scr_vrect.height/2 - 3 + cl_crossy.value;

		glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
		pColor = (unsigned char *) &d_8to24table[(byte) crosshaircolor.value];
		glColor4ubv ( pColor );
		GL_Bind (cs_texture);

		glBegin (GL_QUADS);
		glTexCoord2f (0, 0);
		glVertex2f (x - 4, y - 4);
		glTexCoord2f (1, 0);
		glVertex2f (x+12, y-4);
		glTexCoord2f (1, 1);
		glVertex2f (x+12, y+12);
		glTexCoord2f (0, 1);
		glVertex2f (x - 4, y+12);
		glEnd ();
		
		glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	} else if (crosshair.value)
		Draw_Character (scr_vrect.x + scr_vrect.width/2-4 + cl_crossx.value, 
			scr_vrect.y + scr_vrect.height/2-4 + cl_crossy.value, 
			'+');
}


/*
================
Draw_DebugChar

Draws a single character directly to the upper right corner of the screen.
This is for debugging lockups by drawing different chars in different parts
of the code.
================
*/
void Draw_DebugChar (char num)
{
}

/*
=============
Draw_Pic
=============
*/
void Draw_Pic (int x, int y, qpic_t *pic)
{
	glpic_t			*gl;
	//const float		backdropW = 0.4f;
	//const float		backdropH = 0.4f*((float)pic->height/(float)pic->width);

	if (scrap_dirty)
		Scrap_Upload ();
	gl = (glpic_t *)pic->data;

	/*
	//rww begin
	glColor4f (0,0,0,1);
	glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
	GL_Bind (gl->texnum);
	glBegin (GL_QUADS);
	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x-backdropW, y-backdropH);
	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x+pic->width+backdropW, y-backdropH);
	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x+pic->width+backdropW, y+pic->height+backdropH);
	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x-backdropW, y+pic->height+backdropH);
	glEnd ();
	glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	//rww end
	*/

	glColor4f (1,1,1,1);
	GL_Bind (gl->texnum);
	glBegin (GL_QUADS);
	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);
	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x+pic->width, y);
	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x+pic->width, y+pic->height);
	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y+pic->height);
	glEnd ();
}

/*
=============
Draw_AlphaPic
=============
*/
void Draw_AlphaPic (int x, int y, qpic_t *pic, float alpha)
{
	glpic_t			*gl;

	if (scrap_dirty)
		Scrap_Upload ();
	gl = (glpic_t *)pic->data;
	glDisable(GL_ALPHA_TEST);
	glEnable (GL_BLEND);
//	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glCullFace(GL_FRONT);
	glColor4f (1,1,1,alpha);
	GL_Bind (gl->texnum);
	glBegin (GL_QUADS);
	glTexCoord2f (gl->sl, gl->tl);
	glVertex2f (x, y);
	glTexCoord2f (gl->sh, gl->tl);
	glVertex2f (x+pic->width, y);
	glTexCoord2f (gl->sh, gl->th);
	glVertex2f (x+pic->width, y+pic->height);
	glTexCoord2f (gl->sl, gl->th);
	glVertex2f (x, y+pic->height);
	glEnd ();
	glColor4f (1,1,1,1);
	glEnable(GL_ALPHA_TEST);
	glDisable (GL_BLEND);
}

void Draw_SubPic(int x, int y, qpic_t *pic, int srcx, int srcy, int width, int height)
{
	glpic_t			*gl;
	float newsl, newtl, newsh, newth;
	float oldglwidth, oldglheight;

	if (scrap_dirty)
		Scrap_Upload ();
	gl = (glpic_t *)pic->data;
	
	oldglwidth = gl->sh - gl->sl;
	oldglheight = gl->th - gl->tl;

	newsl = gl->sl + (srcx*oldglwidth)/pic->width;
	newsh = newsl + (width*oldglwidth)/pic->width;

	newtl = gl->tl + (srcy*oldglheight)/pic->height;
	newth = newtl + (height*oldglheight)/pic->height;
	
	glColor4f (1,1,1,1);
	GL_Bind (gl->texnum);
	glBegin (GL_QUADS);
	glTexCoord2f (newsl, newtl);
	glVertex2f (x, y);
	glTexCoord2f (newsh, newtl);
	glVertex2f (x+width, y);
	glTexCoord2f (newsh, newth);
	glVertex2f (x+width, y+height);
	glTexCoord2f (newsl, newth);
	glVertex2f (x, y+height);
	glEnd ();
}

/*
=============
Draw_TransPic
=============
*/
void Draw_TransPic (int x, int y, qpic_t *pic)
{

	if (x < 0 || (unsigned)(x + pic->width) > vid.width || y < 0 ||
		 (unsigned)(y + pic->height) > vid.height)
	{
		Sys_Error ("Draw_TransPic: bad coordinates");
	}
		
	Draw_Pic (x, y, pic);
}


/*
=============
Draw_TransPicTranslate

Only used for the player color selection menu
=============
*/
void Draw_TransPicTranslate (int x, int y, qpic_t *pic, byte *translation)
{
	int				v, u, c;
	unsigned		trans[64*64], *dest;
	byte			*src;
	int				p;

	GL_Bind (translate_texture);

	c = pic->width * pic->height;

	dest = trans;
	for (v=0 ; v<64 ; v++, dest += 64)
	{
		src = &menuplyr_pixels[ ((v*pic->height)>>6) *pic->width];
		for (u=0 ; u<64 ; u++)
		{
			p = src[(u*pic->width)>>6];
			if (p == 255)
				dest[u] = p;
			else
				dest[u] =  d_8to24table[translation[p]];
		}
	}

	glTexImage2D (GL_TEXTURE_2D, 0, gl_alpha_format, 64, 64, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glColor3f (1,1,1);
	glBegin (GL_QUADS);
	glTexCoord2f (0, 0);
	glVertex2f (x, y);
	glTexCoord2f (1, 0);
	glVertex2f (x+pic->width, y);
	glTexCoord2f (1, 1);
	glVertex2f (x+pic->width, y+pic->height);
	glTexCoord2f (0, 1);
	glVertex2f (x, y+pic->height);
	glEnd ();
}

//rww
void Draw_HiResTex(int x, int y, int tex, int w, int h)
{
	float texw = 1.0f;
	float texh = 1.0f;

	GL_Bind (tex);
	glBegin (GL_QUADS);
		glTexCoord2f(0.0f, 0.0f); 
		glVertex2f((float)x, (float)y);
		glTexCoord2f(0.0f, texh);
		glVertex2f((float)x, (float)y+(float)h);
		glTexCoord2f(texw, texh);
		glVertex2f((float)x+(float)w, (float)y+(float)h);
		glTexCoord2f(texw, 0.0f);
		glVertex2f((float)x+(float)w, (float)y);
	glEnd ();
}


/*
================
Draw_ConsoleBackground

================
*/
void Draw_ConsoleBackground (int lines)
{
	char ver[80];
	int x, i;
	int y;
	qboolean blending = false; //rww

	y = (vid.height * 3) >> 2;

	//rww
	if (lines <= y)
	{
		glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );
		glDisable(GL_ALPHA_TEST);
		glEnable (GL_BLEND);
		glCullFace(GL_FRONT);
		glColor4f (1,1,1,(float)(1.2 * lines)/y);
		blending = true;
	}
	else
	{
		glColor4f (1,1,1,1);
	}

	if (hr_conback)
	{ //rww
		Draw_HiResTex(0, lines-vid.height, hr_conback, vid.width, vid.height);
	}
	else
	{
		if (lines > y)
			Draw_Pic(0, lines-vid.height, conback);
		else
			Draw_AlphaPic (0, lines - vid.height, conback, (float)(1.2 * lines)/y);
	}

	// hack the version number directly into the pic
//	y = lines-186;
	y = lines-14;
	if (!cls.download) {
#ifdef __linux__
		sprintf (ver, "LinuxGL (%4.2f) Quake Royale", LINUX_VERSION);
#else
		sprintf (ver, "Quake Royale (%4.2f)", GLQUAKE_VERSION);
#endif
		//x = 116 + vid.conwidth - (strlen(ver)*8 + 11) - (vid.conwidth*8/320)*7;
		x = vid.conwidth - (strlen(ver)*8 + 8);
		for (i=0 ; i<strlen(ver) ; i++)
			Draw_Character (x + i * 8, y, ver[i] | 0x80);
	}

	if (blending)
	{ //rww
		glColor4f (1,1,1,1);
		glEnable(GL_ALPHA_TEST);
		glDisable (GL_BLEND);
		glTexEnvf ( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE );
	}
}


/*
=============
Draw_TileClear

This repeats a 64*64 tile graphic to fill the screen around a sized down
refresh window.
=============
*/
void Draw_TileClear (int x, int y, int w, int h)
{
	glColor3f (1,1,1);
	GL_Bind (*(int *)draw_backtile->data);
	glBegin (GL_QUADS);
	glTexCoord2f (x/64.0, y/64.0);
	glVertex2f (x, y);
	glTexCoord2f ( (x+w)/64.0, y/64.0);
	glVertex2f (x+w, y);
	glTexCoord2f ( (x+w)/64.0, (y+h)/64.0);
	glVertex2f (x+w, y+h);
	glTexCoord2f ( x/64.0, (y+h)/64.0 );
	glVertex2f (x, y+h);
	glEnd ();
}


/*
=============
Draw_Fill

Fills a box of pixels with a single color
=============
*/
void Draw_Fill (int x, int y, int w, int h, int c)
{
	glDisable (GL_TEXTURE_2D);
	glColor3f (host_basepal[c*3]/255.0,
		host_basepal[c*3+1]/255.0,
		host_basepal[c*3+2]/255.0);

	glBegin (GL_QUADS);

	glVertex2f (x,y);
	glVertex2f (x+w, y);
	glVertex2f (x+w, y+h);
	glVertex2f (x, y+h);

	glEnd ();
	glColor3f (1,1,1);
	glEnable (GL_TEXTURE_2D);
}
//=============================================================================

/*
================
Draw_FadeScreen

================
*/
void Draw_FadeScreen (void)
{
	glEnable (GL_BLEND);
	glDisable (GL_TEXTURE_2D);
	glColor4f (0, 0, 0, 0.8);
	glBegin (GL_QUADS);

	glVertex2f (0,0);
	glVertex2f (vid.width, 0);
	glVertex2f (vid.width, vid.height);
	glVertex2f (0, vid.height);

	glEnd ();
	glColor4f (1,1,1,1);
	glEnable (GL_TEXTURE_2D);
	glDisable (GL_BLEND);

	Sbar_Changed();
}

//=============================================================================

/*
================
Draw_BeginDisc

Draws the little blue disc in the corner of the screen.
Call before beginning any disc IO.
================
*/
void Draw_BeginDisc (void)
{
	//rww - don't worry about it.
	/*
	if (!draw_disc)
		return;
	glDrawBuffer  (GL_FRONT);
	Draw_Pic (vid.width - 24, 0, draw_disc);
	glDrawBuffer  (GL_BACK);
	*/
}


/*
================
Draw_EndDisc

Erases the disc icon.
Call after completing any disc IO
================
*/
void Draw_EndDisc (void)
{
}

/*
================
GL_Set2D

Setup as if the screen was 320*200
================
*/
void GL_Set2D (void)
{
	glViewport (glx, gly, glwidth, glheight);

	glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
	if (gl_forceWideAspect.value)
	{ //rww
		glOrtho  (0, vid.width*((float)vid.width/(float)vid.height), vid.height, 0, -99999, 99999);
	}
	else
	{
		glOrtho  (0, vid.width, vid.height, 0, -99999, 99999);
	}

	glMatrixMode(GL_MODELVIEW);
    glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
	glDisable (GL_BLEND);
	glEnable (GL_ALPHA_TEST);
//	glDisable (GL_ALPHA_TEST);

	glColor4f (1,1,1,1);
}

//====================================================================

/*
================
GL_FindTexture
================
*/
int GL_FindTexture (char *identifier)
{
	int		i;
	gltexture_t	*glt;

	for (i=0, glt=gltextures ; i<numgltextures ; i++, glt++)
	{
		if (!strcmp (identifier, glt->identifier))
			return gltextures[i].texnum;
	}

	return -1;
}

/*
================
GL_ResampleTexture
================
*/
void GL_ResampleTexture (unsigned *in, int inwidth, int inheight, unsigned *out,  int outwidth, int outheight)
{
	int		i, j;
	unsigned	*inrow;
	unsigned	frac, fracstep;

	fracstep = inwidth*0x10000/outwidth;
	for (i=0 ; i<outheight ; i++, out += outwidth)
	{
		inrow = in + inwidth*(i*inheight/outheight);
		frac = fracstep >> 1;
		for (j=0 ; j<outwidth ; j+=4)
		{
			out[j] = inrow[frac>>16];
			frac += fracstep;
			out[j+1] = inrow[frac>>16];
			frac += fracstep;
			out[j+2] = inrow[frac>>16];
			frac += fracstep;
			out[j+3] = inrow[frac>>16];
			frac += fracstep;
		}
	}
}

/*
================
GL_Resample8BitTexture -- JACK
================
*/
void GL_Resample8BitTexture (unsigned char *in, int inwidth, int inheight, unsigned char *out,  int outwidth, int outheight)
{
	int		i, j;
	unsigned	char *inrow;
	unsigned	frac, fracstep;

	fracstep = inwidth*0x10000/outwidth;
	for (i=0 ; i<outheight ; i++, out += outwidth)
	{
		inrow = in + inwidth*(i*inheight/outheight);
		frac = fracstep >> 1;
		for (j=0 ; j<outwidth ; j+=4)
		{
			out[j] = inrow[frac>>16];
			frac += fracstep;
			out[j+1] = inrow[frac>>16];
			frac += fracstep;
			out[j+2] = inrow[frac>>16];
			frac += fracstep;
			out[j+3] = inrow[frac>>16];
			frac += fracstep;
		}
	}
}

/*
================
GL_MipMap

Operates in place, quartering the size of the texture
================
*/
void GL_MipMap (byte *in, int width, int height)
{
	int		i, j;
	byte	*out;

	width <<=2;
	height >>= 1;
	out = in;
	for (i=0 ; i<height ; i++, in+=width)
	{
		for (j=0 ; j<width ; j+=8, out+=4, in+=8)
		{
			out[0] = (in[0] + in[4] + in[width+0] + in[width+4])>>2;
			out[1] = (in[1] + in[5] + in[width+1] + in[width+5])>>2;
			out[2] = (in[2] + in[6] + in[width+2] + in[width+6])>>2;
			out[3] = (in[3] + in[7] + in[width+3] + in[width+7])>>2;
		}
	}
}

/*
================
GL_MipMap8Bit

Mipping for 8 bit textures
================
*/
void GL_MipMap8Bit (byte *in, int width, int height)
{
	int		i, j;
	byte	*out;
	unsigned short     r,g,b;
	byte	*at1, *at2, *at3, *at4;

	height >>= 1;
	out = in;
	for (i=0 ; i<height ; i++, in+=width)
		for (j=0 ; j<width ; j+=2, out+=1, in+=2)
		{
			at1 = (byte *) &d_8to24table[in[0]];
			at2 = (byte *) &d_8to24table[in[1]];
			at3 = (byte *) &d_8to24table[in[width+0]];
			at4 = (byte *) &d_8to24table[in[width+1]];

 			r = (at1[0]+at2[0]+at3[0]+at4[0]); r>>=5;
 			g = (at1[1]+at2[1]+at3[1]+at4[1]); g>>=5;
 			b = (at1[2]+at2[2]+at3[2]+at4[2]); b>>=5;

			out[0] = d_15to8table[(r<<0) + (g<<5) + (b<<10)];
		}
}

/*
===============
GL_Upload32
===============
*/
void GL_Upload32 (unsigned *data, int width, int height,  qboolean mipmap, qboolean alpha,
				  unsigned char **out, int *outW, int *outH) //rww
{
	int			samples;
static	unsigned	scaled[1024*1024];//[1024*512];	// [512*256];
	int			scaled_width, scaled_height;

	for (scaled_width = 1 ; scaled_width < width ; scaled_width<<=1)
		;
	for (scaled_height = 1 ; scaled_height < height ; scaled_height<<=1)
		;

	scaled_width >>= (int)gl_picmip.value;
	scaled_height >>= (int)gl_picmip.value;

	if (scaled_width > gl_max_size.value)
		scaled_width = gl_max_size.value;
	if (scaled_height > gl_max_size.value)
		scaled_height = gl_max_size.value;

	if (scaled_width * scaled_height > sizeof(scaled)/4)
		Sys_Error ("GL_LoadTexture: too big");

	samples = alpha ? gl_alpha_format : gl_solid_format;

	if (g_canUseTexComp && gl_texCompression.value && samples == gl_solid_format)
	{ //rww
		samples = 0x83F0;//GL_COMPRESSED_RGB_S3TC_DXT1_EXT;
	}
#if 0
	if (mipmap)
		gluBuild2DMipmaps (GL_TEXTURE_2D, samples, width, height, GL_RGBA, GL_UNSIGNED_BYTE, trans);
	else if (scaled_width == width && scaled_height == height)
		glTexImage2D (GL_TEXTURE_2D, 0, samples, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, trans);
	else
	{
		gluScaleImage (GL_RGBA, width, height, GL_UNSIGNED_BYTE, trans,
			scaled_width, scaled_height, GL_UNSIGNED_BYTE, scaled);
		glTexImage2D (GL_TEXTURE_2D, 0, samples, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
	}
#else
texels += scaled_width * scaled_height;

	if (scaled_width == width && scaled_height == height)
	{
		if (!mipmap)
		{
			glTexImage2D (GL_TEXTURE_2D, 0, samples, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
			if (out) //rww
			{
				*out = (unsigned char *)data;
				*outW = scaled_width;
				*outH = scaled_height;
			}
			goto done;
		}
		memcpy (scaled, data, width*height*4);
	}
	else
		GL_ResampleTexture (data, width, height, scaled, scaled_width, scaled_height);

	glTexImage2D (GL_TEXTURE_2D, 0, samples, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
	if (out) //rww
	{
		if (!mipmap)
		{
			*out = (unsigned char *)scaled;
		}
		else
		{
			static unsigned int t[1024*1024];
			memcpy(t, scaled, (scaled_width*scaled_height)*4);
			*out = (unsigned char *)t;
		}
		*outW = scaled_width;
		*outH = scaled_height;
	}
	if (mipmap)
	{
		int		miplevel;

		miplevel = 0;
		while (scaled_width > 1 || scaled_height > 1)
		{
			GL_MipMap ((byte *)scaled, scaled_width, scaled_height);
			scaled_width >>= 1;
			scaled_height >>= 1;
			if (scaled_width < 1)
				scaled_width = 1;
			if (scaled_height < 1)
				scaled_height = 1;
			miplevel++;
			glTexImage2D (GL_TEXTURE_2D, miplevel, samples, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
		}
	}
done: ;
#endif


	if (mipmap)
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
	}
	else
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_max);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
	}
}

void GL_Upload8_EXT (byte *data, int width, int height,  qboolean mipmap, qboolean alpha) 
{
	int			i, s;
	qboolean	noalpha;
	int			samples;
    static	unsigned char scaled[1024*512];	// [512*256];
	int			scaled_width, scaled_height;

	s = width*height;
	// if there are no transparent pixels, make it a 3 component
	// texture even if it was specified as otherwise
	if (alpha)
	{
		noalpha = true;
		for (i=0 ; i<s ; i++)
		{
			if (data[i] == 255)
				noalpha = false;
		}

		if (alpha && noalpha)
			alpha = false;
	}
	for (scaled_width = 1 ; scaled_width < width ; scaled_width<<=1)
		;
	for (scaled_height = 1 ; scaled_height < height ; scaled_height<<=1)
		;

	scaled_width >>= (int)gl_picmip.value;
	scaled_height >>= (int)gl_picmip.value;

	if (scaled_width > gl_max_size.value)
		scaled_width = gl_max_size.value;
	if (scaled_height > gl_max_size.value)
		scaled_height = gl_max_size.value;

	if (scaled_width * scaled_height > sizeof(scaled))
		Sys_Error ("GL_LoadTexture: too big");

	samples = 1; // alpha ? gl_alpha_format : gl_solid_format;

	texels += scaled_width * scaled_height;

	if (scaled_width == width && scaled_height == height)
	{
		if (!mipmap)
		{
			glTexImage2D (GL_TEXTURE_2D, 0, GL_COLOR_INDEX8_EXT, scaled_width, scaled_height, 0, GL_COLOR_INDEX , GL_UNSIGNED_BYTE, data);
			goto done;
		}
		memcpy (scaled, data, width*height);
	}
	else
		GL_Resample8BitTexture (data, width, height, scaled, scaled_width, scaled_height);

	glTexImage2D (GL_TEXTURE_2D, 0, GL_COLOR_INDEX8_EXT, scaled_width, scaled_height, 0, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, scaled);
	if (mipmap)
	{
		int		miplevel;

		miplevel = 0;
		while (scaled_width > 1 || scaled_height > 1)
		{
			GL_MipMap8Bit ((byte *)scaled, scaled_width, scaled_height);
			scaled_width >>= 1;
			scaled_height >>= 1;
			if (scaled_width < 1)
				scaled_width = 1;
			if (scaled_height < 1)
				scaled_height = 1;
			miplevel++;
			glTexImage2D (GL_TEXTURE_2D, miplevel, GL_COLOR_INDEX8_EXT, scaled_width, scaled_height, 0, GL_COLOR_INDEX, GL_UNSIGNED_BYTE, scaled);
		}
	}
done: ;

	if (mipmap)
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_min);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
	}
	else
	{
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, gl_filter_max);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, gl_filter_max);
	}
}

//rww
static __inline char *GL_LegalizeTexName(char *texname)
{
	static char buf[MAX_OSPATH];
	int i;
	for (i = 0; texname[i] && i < (MAX_OSPATH-1); i++)
	{
		if (texname[i] == '*' ||
			texname[i] == '?' ||
			texname[i] == '\\' ||
			texname[i] == '/')
		{
			buf[i] = '_';
		}
		else
		{
			buf[i] = texname[i];
		}
	}
	buf[i] = 0;

	return buf;
}

//rww
extern png_structp g_png_ptr;
extern png_infop g_info_ptr;

//rww
qboolean GL_UploadPNG(char *filename, int tex, qboolean mips)
{
	unsigned char *dst;
	unsigned int i;
	unsigned int *pixMem;

	int pngRead = read_png(filename);
	if (!pngRead)
	{ //failed to read
		return false;
	}

	if (g_info_ptr->bit_depth != 8 ||
		(g_info_ptr->channels != 3 && g_info_ptr->channels != 4))
	{ //bad format
		close_png();
		return false;
	}

	GL_Bind(tex);

	pixMem = (unsigned int *)malloc(g_info_ptr->width*g_info_ptr->height*4);
	dst = (unsigned char *)pixMem;
	for (i = 0; i < g_info_ptr->height; i++)
	{
		unsigned int j;
		BYTE *rowBuf = (BYTE *)g_info_ptr->row_pointers[i];
		for (j = 0; j < g_info_ptr->rowbytes; j += g_info_ptr->channels)
		{
			dst[0] = rowBuf[j];
			dst[1] = rowBuf[j+1];
			dst[2] = rowBuf[j+2];
			if (g_info_ptr->channels == 4)
			{ //has alpha
				dst[3] = rowBuf[j+3];
			}
			else
			{
				dst[3] = 255;
			}

			dst += 4;
		}
	}

	GL_Upload32(pixMem, g_info_ptr->width, g_info_ptr->height, mips, true, NULL, NULL, NULL);
	free(pixMem);
	close_png();

	return true;
}

//rww - create and upload a normal map
void GL_UploadNormal(char *texname, int tex, int texnorm, unsigned int *data, int width, int height, qboolean mipmap)
{
	char fullpath[MAX_OSPATH];
	unsigned int *normalData;
	int pngRead;

   if (!g_canUseTexFP)
   { //don't upload any normal maps then
	   return;
   }

	sprintf(fullpath, "qdix/nrml/%s.png", GL_LegalizeTexName(texname));

	//bind normal map tex
	GL_Bind(texnorm);

	normalData = (unsigned int *)malloc((width*height)*4);
	pngRead = read_png(fullpath);
	if (pngRead &&
		g_info_ptr->bit_depth == 8 &&
		(g_info_ptr->channels == 3 || g_info_ptr->channels == 4) &&
		g_info_ptr->width == width &&
		g_info_ptr->height == height)
	{ //already have a copy of the file
		unsigned char *dst = (unsigned char *)normalData;
		unsigned int i;
		for (i = 0; i < g_info_ptr->height; i++)
		{
			unsigned int j;
			BYTE *rowBuf = (BYTE *)g_info_ptr->row_pointers[i];
			for (j = 0; j < g_info_ptr->rowbytes; j += g_info_ptr->channels)
			{
				dst[0] = rowBuf[j];
				dst[1] = rowBuf[j+1];
				dst[2] = rowBuf[j+2];
				if (g_info_ptr->channels == 4)
				{ //has alpha
					dst[3] = rowBuf[j+3];
				}
				else
				{
					dst[3] = 255;
				}

				dst += 4;
			}
		}

		close_png();
	}
	else
	{
		if (pngRead)
		{ //had a png, but it was a bad size or format
			close_png();
		}

		//generate and upload the normal map
		GenerateNormalMap((unsigned char *)data, (unsigned char *)normalData, width, height, 4);

		//write it out
		Con_Printf("Writing %s...\n", fullpath);
		GL_Bind(texnorm); //rebind since print function binds a different texture
		png_write((unsigned char *)normalData, width, height, 8, (unsigned char *)fullpath, -1, 1, 0);
	}

	GL_Upload32(normalData, width, height, mipmap, false, NULL, NULL, NULL);
	free(normalData);

	//rebind old texture
	GL_Bind(tex);
}

//rww - returns a pointer to texture memory, filtered or not
unsigned int *GL_FilterTexture(byte *data, int *width, int *height, qboolean mipmap, qboolean alpha,
							   qboolean forceFilter, unsigned int *clrTable)
{
	int			i, s;
	qboolean	noalpha;
	int			p;
	int			transComponents = 6;
	qboolean	hqFilter = true;
	unsigned char *transMem;
	unsigned short *trans;
	unsigned int *transHq;
	int hqLevel = 2;
	int memWidth = *width;
	int memHeight = *height;
	if ((alpha && !clrTable) || !gl_hq4x.value || (!forceFilter && *width > 256 && *height > 256))
	{
		if (!g_loadingPicTex)
		{
			hqFilter = false;
		}
	}

	if (!hqFilter)
	{
		transComponents = 4;
	}
	else
	{
		if (forceFilter || g_loadingPicTex || (*width <= 64 && *height <= 64) || (*width == 320 && *height == 200))
		{ //for super tiny textures and forced hq use 4x (and the console, in a hacked way)
			if (gl_hq4x.value != 2.0f)
			{ //2.0 means only 2x
				hqLevel = 4;
			}
		}
		memWidth *= hqLevel;
		memHeight *= hqLevel;
	}

	transMem = (unsigned char *)malloc((memWidth*memHeight)*transComponents);
	if (hqFilter)
	{
		transHq = (unsigned int *)(transMem);
		trans = (unsigned short *)(transMem+((memWidth*memHeight)*4));
	}
	else
	{
		transHq = (unsigned int *)(transMem);
	}

	s = (*width)*(*height);
	if (s&3)
		Sys_Error ("GL_Upload8: s&3");
	if (clrTable)
	{
		for (i=0 ; i<s ; i++)
		{
			unsigned int c = clrTable[data[i]];

			if (!hqFilter)
			{
				transHq[i] = c;
			}
			else
			{
				if (alpha && !(c>>24))
				{ //alpha clear
					trans[i] = 0;
				}
				else
				{
					unsigned char p;
					p = ((c)&255)/8;
					if (p >= 32)
					{
						p = 31;
					}
					trans[i] = p;
					p = ((c>>8)&255)/4;
					if (p >= 64)
					{
						p = 63;
					}
					trans[i] |= (p<<5);
					p = ((c>>16)&255)/8;
					if (p >= 32)
					{
						p = 31;
					}
					trans[i] |= (p<<11);
				}
			}
		}
	}
	else
	{
		unsigned int *datai = (unsigned int *)data;
		for (i=0 ; i<s ; i++)
		{
			unsigned int c = datai[i];
			if (!hqFilter)
			{
				transHq[i] = c;
				continue;
			}
			if (alpha && !(c>>24))
			{ //alpha clear
				trans[i] = 0;
			}
			else
			{
				unsigned char p;
				p = ((c)&255)/8;
				if (p >= 32)
				{
					p = 31;
				}
				trans[i] = p;
				p = ((c>>8)&255)/4;
				if (p >= 64)
				{
					p = 63;
				}
				trans[i] |= (p<<5);
				p = ((c>>16)&255)/8;
				if (p >= 32)
				{
					p = 31;
				}
				trans[i] |= (p<<11);
			}
		}
	}

	if (hqFilter)
	{
		if (hqLevel == 2)
		{
			hq2x_32((unsigned char *)trans, (unsigned char *)transHq, *width, *height, ((*width)*2)*4);
		}
		else
		{
			hq4x_32((unsigned char *)trans, (unsigned char *)transHq, *width, *height, ((*width)*4)*4);
		}
		*width *= hqLevel;
		*height *= hqLevel;
		if (alpha && clrTable)
		{ //hack to copy original alpha mask back over. todo - render with alpha blend mode and use gradient on edges for smoothness.
			for (i=0 ; i<*height ; i++)
			{
				int j;
				for (j=0 ; j<*width; j++)
				{
					unsigned int lowPix = (j/hqLevel)+((i/hqLevel)*((*width)/hqLevel));
					unsigned int highPix = j+(i*(*width));
					unsigned int c = clrTable[data[lowPix]];
					transHq[highPix] |= ((c>>24)<<24);
				}
			}
		}
	}

	return transHq;
}

/*
===============
GL_Upload8
===============
*/
void GL_Upload8 (gltexture_t *glt, byte *data, int width, int height, qboolean mipmap, qboolean alpha)
{
	//rww - removed old 8-bit nonsense, experimenting with precalculated filtering methods
#if 0
extern qboolean VID_Is8bit();
static	unsigned	trans[640*480];		// FIXME, temporary
	int			i, s;
	qboolean	noalpha;
	int			p;

	s = width*height;
	// if there are no transparent pixels, make it a 3 component
	// texture even if it was specified as otherwise
	if (alpha)
	{
		noalpha = true;
		for (i=0 ; i<s ; i++)
		{
			p = data[i];
			if (p == 255)
				noalpha = false;
			trans[i] = d_8to24table[p];
		}

		if (alpha && noalpha)
			alpha = false;
	}
	else
	{
		if (s&3)
			Sys_Error ("GL_Upload8: s&3");
		for (i=0 ; i<s ; i+=4)
		{
			trans[i] = d_8to24table[data[i]];
			trans[i+1] = d_8to24table[data[i+1]];
			trans[i+2] = d_8to24table[data[i+2]];
			trans[i+3] = d_8to24table[data[i+3]];
		}
	}

	if (VID_Is8bit() && !alpha && (data!=scrap_texels[0])) {
		GL_Upload8_EXT (data, width, height, mipmap, alpha);
		return;
	}

	GL_Upload32 (trans, width, height, mipmap, alpha, NULL, NULL, NULL);
#elif 1
	unsigned int *transHq = GL_FilterTexture(data, &width, &height, mipmap, alpha, false, d_8to24table);
	unsigned char *finalOut;
	int finalW, finalH;
	GL_Upload32 (transHq, width, height, mipmap, alpha, &finalOut, &finalW, &finalH);
	free(transHq);

	if (!alpha && glt && finalOut)
	{ //create/upload corresponding normal map
		GL_UploadNormal(glt->identifier, glt->texnum, glt->normalTexnum, (unsigned int *)finalOut, finalW, finalH, mipmap);
	}
	else if (glt)
	{
		glt->normalTexnum = -1;
	}
#else
	int			i, s;
	int			p;

	int memWidth = width;
	int memHeight = height;
	unsigned char *transMem;
	unsigned short *trans;
	unsigned int *transDelta;
	unsigned short *transSai;
	qboolean saiFilter = true;
	int transComponents = 8;
	if (alpha || !gl_2xsai.value)
	{ //was attempting to handle alphas, but there are too many problems with things like font pages
		saiFilter = false;
	}
	if (saiFilter)
	{
		if (width <= 256 && height <= 256)
		{ //two resamples
			memWidth *= 4;
			memHeight *= 4;
		}
		else
		{
			memWidth *= 2;
			memHeight *= 2;
		}
	}
	else
	{
		transComponents = 4;
	}
	transMem = (unsigned char *)malloc((memWidth*memHeight)*transComponents);
	if (saiFilter)
	{
		trans = (unsigned short *)(transMem);
		transDelta = (unsigned int *)(transMem+((memWidth*memHeight)*2));
		transSai = (unsigned short *)(transMem+((memWidth*memHeight)*2)+((memWidth*memHeight)*4));
	}
	else
	{
		transDelta = (unsigned int *)(transMem);
	}

	s = width*height;
	if (s&3)
		Sys_Error ("GL_Upload8: s&3");
	for (i=0 ; i<s ; i++)
	{
		unsigned int c = d_8to24table[data[i]];

		if (!saiFilter)
		{
			transDelta[i] = c;
		}
		else
		{
			if (alpha && !(c>>24))
			{ //alpha clear
				trans[i] = 0;
			}
			else
			{
				unsigned char p;
				p = ((c)&255)/8;
				if (p >= 32)
				{
					p = 31;
				}
				trans[i] = p;
				p = ((c>>8)&255)/8;
				if (p >= 32)
				{
					p = 31;
				}
				trans[i] |= (p<<5);
				p = ((c>>16)&255)/8;
				if (p >= 32)
				{
					p = 31;
				}
				trans[i] |= (p<<10);
				if (c>>24)
				{ //check for existence of alpha
					trans[i] |= (1<<15);
				}
			}
		}
	}

	if (saiFilter)
	{ //do 2xSaI filter
		Super2xSaI((uint8 *)trans, width*2, (uint8 *)transDelta, (uint8 *)transSai, width*4, width, height);
		width *= 2;
		height *= 2;
		s = width*height;
		if (width <= 512 && height <= 512)
		{ //double sample
			memcpy(trans, transSai, (width*height)*2);
			Super2xSaI((uint8 *)trans, width*2, (uint8 *)transDelta, (uint8 *)transSai, width*4, width, height);
			width *= 2;
			height *= 2;
			s = width*height;
		}

		for (i=0 ; i<s ; i++)
		{
			unsigned char p;
			p = (transSai[i])&31;
			transDelta[i] = (p*8);
			p = (transSai[i]>>5)&31;
			transDelta[i] |= (p*8)<<8;
			p = (transSai[i]>>10)&31;
			transDelta[i] |= (p*8)<<16;
			if (alpha)
			{ //alpha
				if (transSai[i]>>15)
				{
					transDelta[i] |= (255<<24);
				}
			}
		}
	}

	GL_Upload32 (transDelta, width, height, mipmap, alpha, NULL, NULL, NULL);

	free(transMem);
#endif
}

/*
================
GL_LoadTexture
================
*/
int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha, int *normalTex)
{
	int			i;
	gltexture_t	*glt;

	// see if the texture is allready present
	if (identifier[0])
	{
		for (i=0, glt=gltextures ; i<numgltextures ; i++, glt++)
		{
			if (!strcmp (identifier, glt->identifier))
			{
				if (width != glt->width || height != glt->height)
					Sys_Error ("GL_LoadTexture: cache mismatch");
				if (normalTex)
				{ //rww
					*normalTex = gltextures[i].normalTexnum;
				}
				return gltextures[i].texnum;
			}
		}
	}
	else
		glt = &gltextures[numgltextures];
	numgltextures++;

	strcpy (glt->identifier, identifier);
	glt->texnum = texture_extension_number;
	if (normalTex)
	{ //rww
		glt->normalTexnum = texture_extension_number+1;
		*normalTex = glt->normalTexnum;
	}
	glt->width = width;
	glt->height = height;
	glt->mipmap = mipmap;

	GL_Bind(texture_extension_number );

	GL_Upload8 (glt, data, width, height, mipmap, alpha);

	if (normalTex)
	{
		texture_extension_number += 2; //rww
		return texture_extension_number-2;
	}

	texture_extension_number++;
	return texture_extension_number-1;
}

/*
================
GL_LoadPicTexture
================
*/
int GL_LoadPicTexture (qpic_t *pic)
{
	return GL_LoadTexture ("", pic->width, pic->height, pic->data, false, true, NULL);
}

/****************************************/

static GLenum oldtarget = TEXTURE0_SGIS;

void GL_SelectTexture (GLenum target) 
{
	if (!gl_mtexable)
		return;
#ifndef __linux__ // no multitexture under Linux yet
	qglSelectTextureSGIS(target);
#endif
	if (target == oldtarget) 
		return;
	cnttextures[oldtarget-TEXTURE0_SGIS] = currenttexture;
	currenttexture = cnttextures[target-TEXTURE0_SGIS];
	oldtarget = target;
}
