/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// disable data conversion warnings

#pragma warning(disable : 4244)     // MIPS
#pragma warning(disable : 4136)     // X86
#pragma warning(disable : 4051)     // ALPHA
  
#ifdef _WIN32
#include <windows.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

//rww - moved from gl_draw.c
typedef struct
{
	int		texnum;
	int		normalTexnum; //rww
	char	identifier[64];
	int		width, height;
	qboolean	mipmap;
} gltexture_t;

void GL_BeginRendering (int *x, int *y, int *width, int *height);
void GL_EndRendering (void);


// Function prototypes for the Texture Object Extension routines
typedef GLboolean (APIENTRY *ARETEXRESFUNCPTR)(GLsizei, const GLuint *,
                    const GLboolean *);
typedef void (APIENTRY *BINDTEXFUNCPTR)(GLenum, GLuint);
typedef void (APIENTRY *DELTEXFUNCPTR)(GLsizei, const GLuint *);
typedef void (APIENTRY *GENTEXFUNCPTR)(GLsizei, GLuint *);
typedef GLboolean (APIENTRY *ISTEXFUNCPTR)(GLuint);
typedef void (APIENTRY *PRIORTEXFUNCPTR)(GLsizei, const GLuint *,
                    const GLclampf *);
typedef void (APIENTRY *TEXSUBIMAGEPTR)(int, int, int, int, int, int, int, int, void *);

extern	BINDTEXFUNCPTR bindTexFunc;
extern	DELTEXFUNCPTR delTexFunc;
extern	TEXSUBIMAGEPTR TexSubImage2DFunc;

extern	int texture_extension_number;
extern	int		texture_mode;

extern	float	gldepthmin, gldepthmax;

void GL_Upload32 (unsigned *data, int width, int height,  qboolean mipmap, qboolean alpha,
				  unsigned char **out, int *outW, int *outH); //rww
void GL_Upload8 (gltexture_t *glt, byte *data, int width, int height,  qboolean mipmap, qboolean alpha);
void GL_Upload8_EXT (byte *data, int width, int height,  qboolean mipmap, qboolean alpha);
int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha, int *normalTex);
int GL_FindTexture (char *identifier);

//rww begin

#define NUM_ACCUM_TEXTURES 8//4

#define uint32 unsigned int
#define uint16 unsigned short
#define uint8 unsigned char

typedef void (*GLFUNCTION)();
#define GLAPI __stdcall

#define GL_ACTIVE_TEXTURE_ARB               0x84E0
#define GL_CLIENT_ACTIVE_TEXTURE_ARB        0x84E1
#define GL_MAX_TEXTURE_UNITS_ARB            0x84E2
#define GL_TEXTURE0_ARB                     0x84C0
#define GL_TEXTURE1_ARB                     0x84C1
#define GL_TEXTURE2_ARB						0x84C2
#define GL_TEXTURE3_ARB						0x84C3
#define GL_TEXTURE4_ARB						0x84C4
#define GL_TEXTURE5_ARB						0x84C5
#define GL_TEXTURE_3D						0x806F
#define GL_FRAGMENT_PROGRAM_ARB				0x8804
#define GL_VERTEX_PROGRAM_ARB				0x8620
#define GL_PROGRAM_FORMAT_ASCII_ARB			0x8875
#define GL_PROGRAM_ERROR_STRING_ARB			0x8874
#define GL_PROGRAM_ERROR_POSITION_ARB		0x864B
#define GL_TEXTURE_WRAP_R					0x8072
#define GL_CLAMP_TO_EDGE					0x812F
#define GL_CLAMP_TO_BORDER					0x812D
#define GL_TEXTURE_RECTANGLE_NV				0x84F5

//for fbo
#define GL_INVALID_FRAMEBUFFER_OPERATION_EXT 0x0506
#define GL_MAX_RENDERBUFFER_SIZE_EXT      0x84E8
#define GL_FRAMEBUFFER_BINDING_EXT        0x8CA6
#define GL_RENDERBUFFER_BINDING_EXT       0x8CA7
#define GL_FRAMEBUFFER_ATTACHMENT_OBJECT_TYPE_EXT 0x8CD0
#define GL_FRAMEBUFFER_ATTACHMENT_OBJECT_NAME_EXT 0x8CD1
#define GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_LEVEL_EXT 0x8CD2
#define GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_CUBE_MAP_FACE_EXT 0x8CD3
#define GL_FRAMEBUFFER_ATTACHMENT_TEXTURE_3D_ZOFFSET_EXT 0x8CD4
#define GL_FRAMEBUFFER_COMPLETE_EXT       0x8CD5
#define GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT 0x8CD6
#define GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT 0x8CD7
#define GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT 0x8CD8
#define GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT 0x8CD9
#define GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT 0x8CDA
#define GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT 0x8CDB
#define GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT 0x8CDC
#define GL_FRAMEBUFFER_UNSUPPORTED_EXT    0x8CDD
#define GL_MAX_COLOR_ATTACHMENTS_EXT      0x8CDF
#define GL_COLOR_ATTACHMENT0_EXT          0x8CE0
#define GL_COLOR_ATTACHMENT1_EXT          0x8CE1
#define GL_COLOR_ATTACHMENT2_EXT          0x8CE2
#define GL_COLOR_ATTACHMENT3_EXT          0x8CE3
#define GL_COLOR_ATTACHMENT4_EXT          0x8CE4
#define GL_COLOR_ATTACHMENT5_EXT          0x8CE5
#define GL_COLOR_ATTACHMENT6_EXT          0x8CE6
#define GL_COLOR_ATTACHMENT7_EXT          0x8CE7
#define GL_COLOR_ATTACHMENT8_EXT          0x8CE8
#define GL_COLOR_ATTACHMENT9_EXT          0x8CE9
#define GL_COLOR_ATTACHMENT10_EXT         0x8CEA
#define GL_COLOR_ATTACHMENT11_EXT         0x8CEB
#define GL_COLOR_ATTACHMENT12_EXT         0x8CEC
#define GL_COLOR_ATTACHMENT13_EXT         0x8CED
#define GL_COLOR_ATTACHMENT14_EXT         0x8CEE
#define GL_COLOR_ATTACHMENT15_EXT         0x8CEF
#define GL_DEPTH_ATTACHMENT_EXT           0x8D00
#define GL_STENCIL_ATTACHMENT_EXT         0x8D20
#define GL_FRAMEBUFFER_EXT                0x8D40
#define GL_RENDERBUFFER_EXT               0x8D41
#define GL_RENDERBUFFER_WIDTH_EXT         0x8D42
#define GL_RENDERBUFFER_HEIGHT_EXT        0x8D43
#define GL_RENDERBUFFER_INTERNAL_FORMAT_EXT 0x8D44
#define GL_STENCIL_INDEX1_EXT             0x8D46
#define GL_STENCIL_INDEX4_EXT             0x8D47
#define GL_STENCIL_INDEX8_EXT             0x8D48
#define GL_STENCIL_INDEX16_EXT            0x8D49
#define GL_RENDERBUFFER_RED_SIZE_EXT      0x8D50
#define GL_RENDERBUFFER_GREEN_SIZE_EXT    0x8D51
#define GL_RENDERBUFFER_BLUE_SIZE_EXT     0x8D52
#define GL_RENDERBUFFER_ALPHA_SIZE_EXT    0x8D53
#define GL_RENDERBUFFER_DEPTH_SIZE_EXT    0x8D54
#define GL_RENDERBUFFER_STENCIL_SIZE_EXT  0x8D55
#define GL_DEPTH_STENCIL_EXT			  0x84F9
#define GL_UNSIGNED_INT_24_8_EXT		  0x84FA
#define GL_DEPTH24_STENCIL8_EXT			  0x88F0
#define TEXTURE_STENCIL_SIZE_EXT		  0x88F1
#define GL_DEPTH_COMPONENT16              0x81A5
#define GL_DEPTH_COMPONENT24              0x81A6
#define GL_DEPTH_COMPONENT32              0x81A7
#define GL_STENCIL_TEST_TWO_SIDE_EXT	  0x8910
#define GL_ACTIVE_STENCIL_FACE_EXT		  0x8911
#define GL_INCR_WRAP_EXT				  0x8507
#define GL_DECR_WRAP_EXT				  0x8508

#define GL_SAMPLES_PASSED_ARB              0x8914
#define GL_QUERY_COUNTER_BITS_ARB          0x8864
#define GL_CURRENT_QUERY_ARB               0x8865
#define GL_QUERY_RESULT_ARB                0x8866
#define GL_QUERY_RESULT_AVAILABLE_ARB      0x8867

typedef struct glExt_s
{
	BOOL (GLAPI *wglSwapIntervalEXT)(int interval);

	void (GLAPI *glActiveStencilFaceEXT)(GLenum face);

	void (GLAPI *glMultiTexCoord2iARB)(GLenum target, GLint s, GLint t);
	void (GLAPI *glMultiTexCoord2fvARB)(GLenum target, const GLfloat *v);
	void (GLAPI *glMultiTexCoord3fARB)(GLenum target, GLfloat s, GLfloat t, GLfloat r);
	void (GLAPI *glMultiTexCoord3fvARB)(GLenum target, const GLfloat *v);
	void (GLAPI *glActiveTextureARB)(GLenum target);
	void (GLAPI *glClientActiveTextureARB)(GLenum target);

	void (GLAPI *glTexImage3DEXT)(GLenum target, GLint level, GLenum internalformat, GLsizei width, GLsizei height, GLsizei depth, GLint border, GLenum format, GLenum type, const GLvoid *pixels);

	void (GLAPI *glBindProgramNV)(GLenum target, GLuint id);
	void (GLAPI *glLoadProgramNV)(GLenum target, GLuint id, GLsizei len, const GLubyte *program);
	void (GLAPI *glGenProgramsNV)(GLsizei n, GLuint *ids);
	void (GLAPI *glTrackMatrixNV)(GLenum target, GLuint address, GLenum matrix, GLenum transform);
	void (GLAPI *glProgramStringARB)(GLenum target, GLenum format, GLsizei len, const GLvoid *string); 
	void (GLAPI *glBindProgramARB)(GLenum target, GLuint program);
	void (GLAPI *glDeleteProgramsARB)(GLsizei n, const GLuint *programs);
	void (GLAPI *glGenProgramsARB)(GLsizei n, GLuint *programs);

	void (GLAPI *glProgramEnvParameter4dARB)(GLenum target, GLuint index, GLdouble x, GLdouble y, GLdouble z, GLdouble w);
	void (GLAPI *glProgramEnvParameter4dvARB)(GLenum target, GLuint index, const GLdouble *params);
	void (GLAPI *glProgramEnvParameter4fARB)(GLenum target, GLuint index, GLfloat x, GLfloat y, GLfloat z, GLfloat w);
	void (GLAPI *glProgramEnvParameter4fvARB)(GLenum target, GLuint index, const GLfloat *params);

	GLboolean (GLAPI *glIsRenderbufferEXT)(GLuint renderbuffer);
	void (GLAPI *glBindRenderbufferEXT)(GLenum target, GLuint renderbuffer);
	void (GLAPI *glDeleteRenderbuffersEXT)(GLsizei n, const GLuint *renderbuffers);
	void (GLAPI *glGenRenderbuffersEXT)(GLsizei n, GLuint *renderbuffers);
	void (GLAPI *glRenderbufferStorageEXT)(GLenum target, GLenum internalformat, GLsizei width, GLsizei height);
	void (GLAPI *glGetRenderbufferParameterivEXT)(GLenum target, GLenum pname, GLint *params);
	GLboolean (GLAPI *glIsFramebufferEXT)(GLuint framebuffer);
	void (GLAPI *glBindFramebufferEXT)(GLenum target, GLuint framebuffer);
	void (GLAPI *glDeleteFramebuffersEXT)(GLsizei n, const GLuint *framebuffers);
	void (GLAPI *glGenFramebuffersEXT)(GLsizei n, GLuint *framebuffers);
	GLenum (GLAPI *glCheckFramebufferStatusEXT)(GLenum target);
	void (GLAPI *glFramebufferTexture1DEXT)(GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level);
	void (GLAPI *glFramebufferTexture2DEXT)(GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level);
	void (GLAPI *glFramebufferTexture3DEXT)(GLenum target, GLenum attachment, GLenum textarget, GLuint texture, GLint level, GLint zoffset);
	void (GLAPI *glFramebufferRenderbufferEXT)(GLenum target, GLenum attachment, GLenum renderbuffertarget, GLuint renderbuffer);
	void (GLAPI *glGetFramebufferAttachmentParameterivEXT)(GLenum target, GLenum attachment, GLenum pname, GLint *params);
	void (GLAPI *glGenerateMipmapEXT)(GLenum target);

	void (GLAPI *glGenQueriesARB)(GLsizei n, GLuint *ids);
	void (GLAPI *glDeleteQueriesARB)(GLsizei n, const GLuint *ids);
	GLboolean (GLAPI *glIsQueryARB)(GLuint id);
	void (GLAPI *glBeginQueryARB)(GLenum target, GLuint id);
	void (GLAPI *glEndQueryARB)(GLenum target);
	void (GLAPI *glGetQueryivARB)(GLenum target, GLenum pname, GLint *params);
	void (GLAPI *glGetQueryObjectivARB)(GLuint id, GLenum pname, GLint *params);
	void (GLAPI *glGetQueryObjectuivARB)(GLuint id, GLenum pname, GLuint *params);
} glExt_t;
extern glExt_t g_glExt;

extern cvar_t gl_normalMaps;
extern cvar_t gl_bloom;
extern cvar_t gl_dynamicMapLights;
extern cvar_t gl_ambientShadows;
extern cvar_t gl_qualityFoliage;
extern cvar_t gl_lowQualityBloom;
extern qboolean g_canUseTexFP;
extern qboolean g_canUseFBO;
extern qboolean g_canUseTexRect;
extern qboolean g_twoSidedStencil;
extern qboolean g_canUseOccQuery;
extern int g_fpLightAmbientH;
extern int g_fpLightModelAmbientH;
extern int g_fpLightRDMAmbientH;
extern int g_fpLightAmbientDLightH;
extern int g_fpLightWaterH;
extern int g_fpSpecPassH;
extern int g_fpBloomFragH;
extern int g_fpBloomVertH;

extern int screen_texture;

extern viewFrustum_t g_curFrustum;
extern float g_projectionMatrix[16];

extern dlight_t *g_liveLights[MAX_DLIGHTS];
extern int g_liveLightsIdx[MAX_DLIGHTS];
extern int g_numLiveLights;

extern qboolean g_useSSE;
extern qboolean g_useSSE2;

//r_2xsai.c
int Init_2xSaI(uint32 BitFormat);
void Super2xSaI(uint8 *srcPtr, uint32 srcPitch,
	     uint8 *deltaPtr,
	     uint8 *dstPtr, uint32 dstPitch, int width, int height);

//r_hq2x.c
void hq2x_InitLUTs(void);
void hq2x_32( unsigned char * pIn, unsigned char * pOut, int Xres, int Yres, int BpL );

//r_hq4x.c
void hq4x_InitLUTs(void);
void hq4x_32( unsigned char * pIn, unsigned char * pOut, int Xres, int Yres, int BpL );

//r_nmapgen.c
int GenerateNormalMap(unsigned char *src, unsigned char *dst, int dstWidth, int dstHeight, int dstBpp);

//gl_draw.c
void GL_UploadNormal(char *texname, int tex, int texnorm, unsigned int *data, int width, int height, qboolean mipmap);
unsigned int *GL_FilterTexture(byte *data, int *width, int *height, qboolean mipmap, qboolean alpha,
							   qboolean forceFilter, unsigned int *clrTable);

//rww end

typedef struct
{
	float	x, y, z;
	float	s, t;
	float	r, g, b;
} glvert_t;

extern glvert_t glv;

extern	int glx, gly, glwidth, glheight;

#ifdef _WIN32
extern	PROC glArrayElementEXT;
extern	PROC glColorPointerEXT;
extern	PROC glTexturePointerEXT;
extern	PROC glVertexPointerEXT;
#endif

// r_local.h -- private refresh defs

#define ALIAS_BASE_SIZE_RATIO		(1.0 / 11.0)
					// normalizing factor so player model works out to about
					//  1 pixel per triangle
#define	MAX_LBM_HEIGHT		480

#define TILE_SIZE		128		// size of textures generated by R_GenTiledSurf

#define SKYSHIFT		7
#define	SKYSIZE			(1 << SKYSHIFT)
#define SKYMASK			(SKYSIZE - 1)

#define BACKFACE_EPSILON	0.01


void R_TimeRefresh_f (void);
void R_ReadPointFile_f (void);
texture_t *R_TextureAnimation (texture_t *base);

typedef struct surfcache_s
{
	struct surfcache_s	*next;
	struct surfcache_s 	**owner;		// NULL is an empty chunk of memory
	int					lightadj[MAXLIGHTMAPS]; // checked for strobe flush
	int					dlight;
	int					size;		// including header
	unsigned			width;
	unsigned			height;		// DEBUG only needed for debug
	float				mipscale;
	struct texture_s	*texture;	// checked for animating textures
	byte				data[4];	// width*height elements
} surfcache_t;


typedef struct
{
	pixel_t		*surfdat;	// destination for generated surface
	int			rowbytes;	// destination logical width in bytes
	msurface_t	*surf;		// description for surface to generate
	fixed8_t	lightadj[MAXLIGHTMAPS];
							// adjust for lightmap levels for dynamic lighting
	texture_t	*texture;	// corrected for animating textures
	int			surfmip;	// mipmapped ratio of surface texels / world pixels
	int			surfwidth;	// in mipmapped texels
	int			surfheight;	// in mipmapped texels
} drawsurf_t;


typedef enum {
	pt_static, pt_grav, pt_slowgrav, pt_fire, pt_explode, pt_explode2, pt_blob, pt_blob2
} ptype_t;

// !!! if this is changed, it must be changed in d_ifacea.h too !!!
typedef struct particle_s
{
// driver-usable fields
	vec3_t		org;
	float		color;
// drivers never touch the following fields
	struct particle_s	*next;
	vec3_t		vel;
	float		ramp;
	float		die;
	ptype_t		type;
} particle_t;


//====================================================


extern	entity_t	r_worldentity;
extern	qboolean	r_cache_thrash;		// compatability
extern	vec3_t		modelorg, r_entorigin;
extern	entity_t	*currententity;
extern	int			r_visframecount;	// ??? what difs?
extern	int			r_framecount;
extern	mplane_t	frustum[4];
extern	int		c_brush_polys, c_alias_polys;


//
// view origin
//
extern	vec3_t	vup;
extern	vec3_t	vpn;
extern	vec3_t	vright;
extern	vec3_t	r_origin;

//
// screen size info
//
extern	refdef_t	r_refdef;
extern	mleaf_t		*r_viewleaf, *r_oldviewleaf;
extern	texture_t	*r_notexture_mip;
extern	int		d_lightstylevalue[256];	// 8.8 fraction of base light value

extern	qboolean	envmap;
extern	int	currenttexture;
extern	int	cnttextures[2];
extern	int	particletexture;
extern	int	netgraphtexture;	// netgraph texture
extern	int	playertextures;

extern	int	skytexturenum;		// index in cl.loadmodel, not gl texture object

extern	cvar_t	r_norefresh;
extern	cvar_t	r_drawentities;
extern	cvar_t	r_drawworld;
extern	cvar_t	r_drawviewmodel;
extern	cvar_t	r_speeds;
extern	cvar_t	r_waterwarp;
extern	cvar_t	r_fullbright;
extern	cvar_t	r_lightmap;
extern	cvar_t	r_shadows;
extern	cvar_t	r_mirroralpha;
extern	cvar_t	r_wateralpha;
extern	cvar_t	r_fancywater; //rww
extern	cvar_t	r_dynamic;
extern	cvar_t	r_novis;
extern	cvar_t	r_netgraph;

extern	cvar_t	gl_clear;
extern	cvar_t	gl_cull;
extern	cvar_t	gl_poly;
extern	cvar_t	gl_texsort;
extern	cvar_t	gl_smoothmodels;
extern	cvar_t	gl_affinemodels;
extern	cvar_t	gl_polyblend;
extern	cvar_t	gl_keeptjunctions;
extern	cvar_t	gl_reporttjunctions;
extern	cvar_t	gl_flashblend;
extern	cvar_t	gl_nocolors;
extern	cvar_t	gl_finish;
extern	cvar_t	gl_fognear; //rww
extern	cvar_t	gl_fogfar; //rww
extern	cvar_t	gl_fog_r; //rww
extern	cvar_t	gl_fog_g; //rww
extern	cvar_t	gl_fog_b; //rww


extern	int		gl_lightmap_format;
extern	int		gl_solid_format;
extern	int		gl_alpha_format;

extern	cvar_t	gl_max_size;
extern	cvar_t	gl_playermip;

extern	int			mirrortexturenum;	// quake texturenum, not gltexturenum
extern	qboolean	mirror;
extern	mplane_t	*mirror_plane;

extern	float	r_world_matrix[16];

extern	const char *gl_vendor;
extern	const char *gl_renderer;
extern	const char *gl_version;
extern	const char *gl_extensions;

void R_TranslatePlayerSkin (int playernum);
void GL_Bind (int texnum);

// Multitexture
#define    TEXTURE0_SGIS				0x835E
#define    TEXTURE1_SGIS				0x835F

#ifdef _WIN32
typedef void (APIENTRY *lpMTexFUNC) (GLenum, GLfloat, GLfloat);
typedef void (APIENTRY *lpSelTexFUNC) (GLenum);
extern lpMTexFUNC qglMTexCoord2fSGIS;
extern lpSelTexFUNC qglSelectTextureSGIS;
#endif

extern qboolean gl_mtexable;

void GL_DisableMultitexture(void);
void GL_EnableMultitexture(void);

//
// gl_warp.c
//
void GL_SubdivideSurface (msurface_t *fa);
void EmitBothSkyLayers (msurface_t *fa);
void EmitWaterPolys (msurface_t *fa);
void EmitSkyPolys (msurface_t *fa);
void R_DrawSkyChain (msurface_t *s);

//
// gl_draw.c
//
int GL_LoadPicTexture (qpic_t *pic);
void GL_Set2D (void);

//
// gl_rmain.c
//
qboolean R_CullBox (vec3_t mins, vec3_t maxs);
void R_RotateForEntity (entity_t *e);

//
// gl_rlight.c
//
void R_MarkLights (dlight_t *light, unsigned int bit, unsigned int bitHigh, mnode_t *node); //rww
void R_AnimateLight (void);
void R_RenderDlights (void);
int R_LightPoint (vec3_t p);

//
// gl_refrag.c
//
void R_StoreEfrags (efrag_t **ppefrag);

//
// gl_mesh.c
//
void GL_MakeAliasModelDisplayLists (model_t *m, aliashdr_t *hdr);

//
// gl_rsurf.c
//
void R_DrawBrushModel (entity_t *e, int lightNum); //rww - light
void R_DrawWorld (void);
void GL_BuildLightmaps (void);

//
// gl_ngraph.c
//
void R_NetGraph (void);

